# pylint: disable=C0111,W0611,W0614
from sys import modules

import ansi.colour  # lgtm [py/import-and-import-from]
from ansi.colour import bg, fg, fx, rgb

modules["ansi.color"] = ansi.colour

__all__ = ["fg", "bg", "fx", "rgb"]



import typing

import vim_quest_cli.deps.ansi.color as ansi_color

from vim_quest_cli.deps.is_js import IS_JS
from vim_quest_cli.interface.mapping import input_code_to_enum, KeyNames
from vim_quest_cli.mode.mode_interface import Mode

if not IS_JS:
    import curses
    from curses import window as curses_window
else:
    curses_window = typing.Any

from vim_quest_cli.deps.with_stdin_raw import with_stdin_raw
from vim_quest_cli.engine.engine_interface import EngineState, EngineInterface
from vim_quest_cli.interface.curses_window import Window
from vim_quest_cli.view.view import (
    ViewFactory,
    ViewData,
    CharacterView,
    RESET_STYLE,
    FgColor,
)


class TerminalUI:
    def __init__(
        self,
        mode: Mode,
        unicode_supported=True,
    ):
        self._mode: Mode = mode
        self._unicode_supported = unicode_supported
        self._screen_size_has_been_initialized = False

    def _assure_screen_size_present(self, stdscr):
        if self._screen_size_has_been_initialized:
            return
        self._screen_size_has_been_initialized = True
        screen_rows, screen_cols = stdscr.getmaxyx()
        self._state = self._state.copy(
            screen_cols=screen_cols,
            screen_rows=screen_rows - 1,  # Leave space for the status
        )

    def loop_one_step(self, input_str: str, stdscr: curses_window):
        # self._assure_screen_size_present(stdscr)

        if not input_str:
            view = self._mode.feedkey()
            self._refresh_state(view, stdscr)
            return

        key_name = input_code_to_enum(input_str)
        if key_name in {"q", "Q", KeyNames.CTRL_C}:
            raise StopIteration()

        view = self._mode.feedkey([key_name])

        self._refresh_state(view, stdscr)

    def _clean_before_leaving(self, stdscr: curses_window):
        stdscr.clear()
        stdscr.move(0, 0)
        stdscr.refresh()

    def start_with_window(self, stdscr: Window):

        try:
            self.loop_one_step(None, stdscr)
            while True:
                input_str = stdscr.getkey()
                self.loop_one_step(input_str, stdscr)
        except StopIteration:
            self._clean_before_leaving(stdscr)
            return

    def start_ansi(self):
        with with_stdin_raw():
            self.start_with_window(Window())

    def start_curses(self):
        curses.wrapper(self.start_with_window)

    def _refresh_state(self, view: ViewData, stdscr: Window):
        stdscr.clear()  # TODO there should be a way to avoid using that, it makes the screen blink.

        stdscr._write("\033[?25l")  # hide cursor

        def process_character(c: CharacterView) -> str:
            fg_color = c.fg_color.value if c.fg_color is not None else ""
            bg_color = c.bg_color.value if c.bg_color is not None else ""
            return "".join([fg_color, bg_color, c.ascii, RESET_STYLE])

        def process_line(l: typing.Iterable[CharacterView]) -> str:
            return "".join(process_character(c) for c in l)

        to_show = "\n".join(process_line(l) for l in view.chars)
        stdscr.addstr(0, 0, to_show)

        stdscr.move(view.cursor.line - 1, view.cursor.column - 1)

        stdscr._write("\033[?25h")  # Show cursor
        stdscr.refresh()



import binascii
from typing import Any, Union

from ansi.sequence import osc


def _b64(data: Union[bytes, str]) -> str:
    if not isinstance(data, bytes):
        data = data.encode()
    return binascii.b2a_base64(data, newline=False).decode()


notification = osc(9)


def badge(msg: str) -> str:
    return osc(1337)(SetBadgeFormat=_b64(msg))


def image(data: str, name: str = "unnamed", **kwargs: Any) -> str:
    kwargs["inline"] = 1
    return (
        osc(1337)("File=name=%s" % _b64(name), **kwargs)[:-1] + ":" + _b64(data) + "\a"
    )


def cursor(shape: str) -> str:
    return osc(1337)(CursorShape=shape)


def cursorguide(on: Union[bool, str]) -> str:
    if isinstance(on, bool):
        on = ["no", "yes"][on]
    return osc(1337)(HighlightCursorLine=on)


def attention(on: Union[bool, str]) -> str:
    if isinstance(on, bool):
        on = ["no", "yes"][on]
    return osc(1337)(RequestAttention=on)



# First have a method that take a suite of
import dataclasses
import enum
import itertools
from typing import List


@dataclasses.dataclass(frozen=True)
class CursorPos:
    line: int = 1  # 1 based. The first line is 1. Even without any char in the line.
    col: int = 1  # 1 based. The first element of the line is 1. Even without lines.
    col_want: int = 1  # Same as columns.

    def copy(self, **changes):
        return dataclasses.replace(self, **changes)


DEFAULT_CURSOR_POS = CursorPos()


@dataclasses.dataclass(frozen=True)
class EngineState:
    cursor: CursorPos = DEFAULT_CURSOR_POS

    # Todo later
    # screen_rows: int = 2
    # screen_cols: int = 2
    # screen_xpos: int = 0
    # screen_ypos: int = 0

    buffer: List[str] = dataclasses.field(default_factory=list)
    # paste: str = ""
    # command: str = ""
    # search: str = ""
    command: List[str] = dataclasses.field(default_factory=list)

    # should_quit: bool = False

    # action: None = None  # Placeholder for any other kind of action outside the status

    def copy(self, **changes):
        return dataclasses.replace(self, **changes)

    def with_added_command(self, command: List[str]):
        if not command:
            return self
        return self.copy(command=tuple(itertools.chain(self.command, command)))


class EngineInterface:
    def process(self, state: EngineState) -> EngineState:
        raise NotImplemented



"""
Return a text code for every input.
"""
import select
import sys
import tty


def test_stdin(timeout=0.5):
    fno = sys.stdin.fileno()
    tty.setraw(fno)

    res = sys.stdin.read(1)  # Blocking
    rfds, _, _ = select.select([fno], [], [], timeout)

    while sys.stdin.fileno() in rfds or sys.stdin in rfds:
        res += sys.stdin.read(1)
        rfds, _, _ = select.select([sys.stdin.fileno()], [], [], 0)

    return res


def main():
    print("Starting stdin tests")
    while True:
        v = test_stdin()
        print(f"{v=!r}")
        if v == "q":
            break


if __name__ == "__main__":
    main()



import sys
import sys
from importlib import _bootstrap
from importlib.abc import MetaPathFinder
from types import ModuleType
from typing import Optional, Dict

from dataclasses import dataclass

from vim_quest_cli.deps.project2singleFile_export import ModuleDef, ModuleImporter

# True for both nodejs and browser.
IS_JS = sys.platform == "emscripten"

MODULES_TO_FAKE = ("termios", "curses", "curses.ascii", "importlib_resources")


def _module_def_from_name(module_name: str) -> ModuleDef:
    return ModuleDef(
        name=module_name,
        source="# Fake module",
        is_package=True,
        file=module_name.replace(".", "/") + ".py",
    )


def _create_fake_modules(modules_names=MODULES_TO_FAKE) -> None:
    modules = [_module_def_from_name(name) for name in modules_names]
    ModuleImporter.install(modules)


if IS_JS:
    sys.meta_path = sys.meta_path[:-1]
    _create_fake_modules()



# pylint: disable=C0103,R0903

from ansi.colour.base import Graphic


# ECMA-048 standard names
reset = Graphic("0")
bold = Graphic("1")
faint = Graphic("2")
italic = Graphic("3")
underline = Graphic("4")
blink_slow = Graphic("5")
blink = Graphic("6")
negative = Graphic("7")
conceal = Graphic("8")
crossed_out = Graphic("9")
font_reset = Graphic("10")
font_1 = Graphic("11")
font_2 = Graphic("12")
font_3 = Graphic("13")
font_4 = Graphic("14")
font_5 = Graphic("15")
font_6 = Graphic("16")
font_7 = Graphic("17")
font_8 = Graphic("18")
font_9 = Graphic("19")
fraktur = Graphic("20")
gothic = Graphic("20")
underline_double = Graphic("21")
normal = Graphic("22")
not_italic = Graphic("23")
not_fraktur = Graphic("23")
not_gothic = Graphic("23")
not_underline = Graphic("24")
steady = Graphic("25")
positive = Graphic("27")
reveal = Graphic("28")
framed = Graphic("51")
encircled = Graphic("52")
overlined = Graphic("53")
not_framed = Graphic("54")
not_encircled = Graphic("54")
not_overlined = Graphic("55")

# Convenience wrappers
inverse = negative
bright = bold
not_blink = steady
blink_off = steady



from typing import Iterable

from vim_quest_cli.engine.engine_interface import EngineState
from vim_quest_cli.mode.mode_interface import Mode
from vim_quest_cli.view.view import ViewData, FgColor


class LabyrinthMode(Mode):
    @staticmethod
    def _char_under_cursor(state: EngineState):
        return state.buffer[state.cursor.line - 1][state.cursor.col - 1]

    def state_change(
        self, state_init: EngineState, state_end: EngineState
    ) -> EngineState:
        if LabyrinthMode._char_under_cursor(state_end) == "#":
            return state_init.copy(command=state_end.command)
        return state_end

    def change_view(self, view: ViewData):
        for line in view.chars:
            for character in line:
                if character.ascii == "#":
                    character.fg_color = FgColor.red

        return view



"""
Common process around the commands and what format they have.

"""
import enum
import string

ALL_RAW = set(string.printable)


@enum.Enum
class KeyEnum:
    LEFT = "<Left>"
    RIGHT = "<Right>"
    UP = "<Up>"
    DOWN = "<Down>"



from dataclasses import dataclass
from typing import List

from vim_quest_cli.engine.engine_interface import EngineState
from vim_quest_cli.old.status import VimGlobalStatus, SpecialChar


def assign_screen_size(
    state: VimGlobalStatus, nb_rows: int, nb_cols: int
) -> VimGlobalStatus:
    return state.copy(
        screen_cols=nb_rows,
        screen_rows=nb_cols - 1,  # Leave space for the status
    )


@dataclass(frozen=True)
class ScreenSize:
    rows: int
    cols: int


@dataclass(frozen=True)
class CursorPos:
    line: int
    col: int


@dataclass(frozen=True)
class ScreenData:
    screen_viewport: List[str]
    status_line: str
    cursor_pos: CursorPos


@dataclass(frozen=True)
class ScreenPositionState:
    xpos: int = 1
    ypos: int = 1
    cols: int = 80
    rows: int = 24


def get_status_line(state: EngineState) -> str:
    return f"{state.command!r}  pos: {state.cursor.line},{state.cursor.col}"


def get_screen_viewport(
    state: EngineState, screen_pos: ScreenPositionState, unicode_supported=True
) -> List[str]:
    xpos, nbcols = screen_pos.xpos, screen_pos.cols
    ypos, nbrows = screen_pos.ypos, screen_pos.rows

    x_start, x_end = xpos, xpos + nbcols
    y_start, y_end = ypos, ypos + nbrows

    relevant_lines = state.buffer[y_start:y_end]
    res = [l[x_start:x_end] for l in relevant_lines]

    if not unicode_supported:

        def _replace_unicode_line(l: str, replace_by="??"):
            return "".join((replace_by if SpecialChar.is_emoji(c) else c) for c in l)

        res = [_replace_unicode_line(l) for l in relevant_lines]

    return res


def get_cursor_pos_on_screen(
    state: EngineState, screen_pos: ScreenPositionState
) -> CursorPos:
    return CursorPos(
        line=state.cursor.line - screen_pos.ypos - 1,
        col=state.cursor.col - screen_pos.xpos - 1,
    )


def get_screen_data(state: EngineState, unicode_supported=True) -> ScreenData:
    screen_pos = ScreenPositionState()  # Extract from the engine state for later
    return ScreenData(
        status_line=get_status_line(state),
        screen_viewport=get_screen_viewport(state, screen_pos, unicode_supported),
        # cursor_pos=get_cursor_pos_on_screen(state, screen_pos),
        cursor_pos=CursorPos(-1, -1),
    )



# pylint: disable=C0103,R0903

from ansi.colour.base import Graphic
from ansi.colour.fx import bold


# ECMA-048 standard names
black = Graphic("30")
red = Graphic("31")
green = Graphic("32")
yellow = Graphic("33")
blue = Graphic("34")
magenta = Graphic("35")
cyan = Graphic("36")
white = Graphic("37")
default = Graphic("39")

# ECMA-048 bold variants
boldblack = bold + black
boldred = bold + red
boldgreen = bold + green
boldyellow = bold + yellow
boldblue = bold + blue
boldmagenta = bold + magenta
boldcyan = bold + cyan
boldwhite = bold + white

# High intensity variants
brightblack = Graphic("90")
brightred = Graphic("91")
brightgreen = Graphic("92")
brightyellow = Graphic("93")
brightblue = Graphic("94")
brightmagenta = Graphic("95")
brightcyan = Graphic("96")
brightwhite = Graphic("97")

# Convenience wrappers
brown = yellow  # Not in ANSI/ECMA-048 standard
grey = white  # Not in ANSI/ECMA-048 standard
gray = white  # US English
darkgrey = boldblack
darkgray = boldblack  # US English
brightbrown = boldyellow  # Not in ANSI/ECMA-048 standard
brightgrey = boldwhite  # Not in ANSI/ECMA-048 standard
brightgray = boldwhite  # Us English

# 8 bit and 24 bit colors colors
palette = lambda colour: Graphic("38;5;%s" % colour)
truecolor = lambda r, g, b: Graphic("38;2;%s;%s;%s" % (r, g, b))



# https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797#file-ansi-md
import enum
import os
import re
import select
import sys
from typing import Tuple

from vim_quest_cli.deps.is_js import IS_JS

if not IS_JS:
    import curses

from vim_quest_cli.deps.ansi import cursor as ansi_cursor
from vim_quest_cli.deps.ansi.sequence import sequence


# TODO for cursor position :
# https://stackoverflow.com/questions/46651602/determine-the-terminal-cursor-position-with-an-ansi-sequence-in-python-3


def get_char(file, timeout=0):
    if select.select([file], [], [], timeout) == ([file], [], []):
        print("ready for reading")
        return file.read(1)
    return ""


class Ansi:
    # Escape sequence = "\033"
    CLEAR_ENTIRE_SCREEN = "\033[2J"
    RESET = "\033[c"
    MOVE_CURSOR = "\033[{line};{column}H"
    REQUEST_CURSOR_POSITION = "\033[6n"
    REQUEST_CURSOR_POSITION_2 = sequence("6n")()
    assert REQUEST_CURSOR_POSITION == REQUEST_CURSOR_POSITION_2
    HIDE_CURSOR = "\033[?25l"
    SHOW_CURSOR = "\033[?25h"

    @staticmethod
    def cursor_goto(row: int, col: int):
        return ansi_cursor.goto(row, col)

    # Reading sequence
    CURSOR_POSITION_RESPONSE = re.compile(r"\033\[(?P<line>[0-9]*);(?P<column>[0-9]*)R")

    def getmaxyx(self):
        return curses.wrapper(lambda stdscr: stdscr.getmaxyx())


def stdout_write(s):
    sys.stdout.write(s)
    sys.stdout.flush()


def reading_stdin(blocking=True, timeout=0.5):
    import tty

    # Not needed anymore
    # tty.setraw(sys.stdin)
    return sys.stdin.read(1)
    # return sys.stdin.read(1)  # In the meantime

    # tty.setraw(sys.stdin.fileno())
    if blocking:
        res = sys.stdin.read(1)  # Blocking
    else:
        res = ""

    tty.setraw(sys.stdin.fileno())

    rfds, _, _ = select.select([sys.stdin.fileno()], [], [], timeout)
    while sys.stdin.fileno() in rfds or sys.stdin in rfds:
        res += sys.stdin.read(1)
        rfds, _, _ = select.select([sys.stdin.fileno()], [], [], 0)

    return res


class InputKeysNames(enum.Enum):
    pass


def reset_terminal():
    stdout_write(Ansi.RESET)


def interpret_input(s):
    char = ord(s)
    if char == 3:  # Ctr-C
        raise KeyboardInterrupt()

    if 32 <= char <= 126:  # Regular characters
        return chr(char)

    if char in {10, 13}:  # Enter
        return "ENTER"  # TODO: check

    if char == 27:
        # It's also the ESC sequence
        return "KEY_ESC"  # In the meantime

        # if not self._has_stdin_more(timeout=0):
        #    return 'KEY_ESC'

        # next1, next2 = ord(sys.stdin.read(1)), ord(sys.stdin.read(1))
        # command = (char, next1, next2)
        # if command == (27, 91, 68):
        #    return "KEY_LEFT"
        # if command == (27, 91, 67):
        #    return "KEY_RIGHT"
        # if command == (27, 91, 66):
        #    return "KEY_DOWN"
        # if command == (27, 91, 65):
        #    return "KEY_UP"
        return f"{command=!r}"


class Window:
    """Ansi window that duck-type the curses window."""

    def __init__(self, reading_stdin_fct=reading_stdin, stdout_write_fct=stdout_write):
        self._reading_stdin_fct = reading_stdin_fct
        self._stdout_write_fct = stdout_write_fct

        self._stdout_written = []

    def _write(self, s):
        self._stdout_written.append(s)

    def _has_stdin_more(self, timeout=0.1):
        return select.select([self._stdin], [], [], 0) == ([self._stdin], [], [])

    def clear(self) -> None:
        self._write(Ansi.CLEAR_ENTIRE_SCREEN)

    def getmaxyx(self) -> Tuple[int, int]:
        return (24, 80)
        if os.getenv("VIMQUEST_MAXYX"):
            max_str = os.getenv("VIMQUEST_MAXYX")
            max_y_str, max_x_str = max_str.split("-")
            return int(max_y_str), int(max_x_str)

        return curses.wrapper(lambda stdscr: stdscr.getmaxyx())

    def addstr(self, row: int, col: int, s: str):
        self._write(Ansi.cursor_goto(row, col))
        self._write(s.replace("\n", "\n\r"))

    def move(self, row: int, col: int) -> None:
        row += 1  # Values are base 1 in ansi
        col += 1
        self._write(Ansi.cursor_goto(row, col))

    def refresh(self) -> None:
        self._stdout_write_fct("".join(self._stdout_written))
        self._stdout_written.clear()

    def getkey(self) -> str:
        try:
            return interpret_input(self._reading_stdin_fct())
        except Exception as _:
            reset_terminal()
            raise



# pylint: disable=C0103,R0903
from typing import TypeVar, Union, cast

from ansi.sequence import sequence

__all__ = ["Graphic"]

T = TypeVar("T", str, "Graphic")


class Graphic(object):
    """
    Compose a Select Graphic Rendition (SGR) ANSI escape sequence.
    """

    def __init__(self, *values: Union[str, "Graphic"]) -> None:
        self.values = values
        self.sequence = sequence("m", fields=-1)(*values)

    def __add__(self, their: T) -> T:
        if isinstance(their, str):
            # For some reason mypy does not understand that T is string in this case
            return cast(T, self.sequence + their)
        elif isinstance(their, Graphic):
            return Graphic(*(self.values + their.values))
        raise TypeError(
            "You can only add strings or strings decorated with escape sequences together."
        )

    def __call__(self, text: str, reset: bool = True) -> str:
        result = self.sequence + text
        if reset:
            result += str(Graphic("0"))
        return result

    def __str__(self) -> str:
        return self.sequence



"""
When reading from stdin, if we want to be efficient, we need to read without cache and fast.
That's why we want to read stdin in a raw mode, but this messes with the stdin and the terminal doesn't like that.
"""
import contextlib
import subprocess
import sys

from vim_quest_cli.deps.is_js import IS_JS

if not IS_JS:
    import termios
    import tty


@contextlib.contextmanager
def with_stdin_raw():
    try:
        with contextlib.suppress(termios.error):
            tty.setraw(sys.stdin)
        yield
    finally:
        with contextlib.suppress(termios.error):
            tty.setraw(sys.stdin)  # Doesn't really works.
        # Should have a way of doing that in python, but maybe for later.
        subprocess.check_call(["stty", "sane"])



# First have a method that take a suite of
import dataclasses
import enum
from typing import List

# import emoji
# from emoji import EMOJI_UNICODE_ENGLISH, UNICODE_EMOJI_ALIAS_ENGLISH
EMOJI_UNICODE_ENGLISH = {
    ":brick:": "\U0001F9F1",
}

UNICODE_EMOJI_ALIAS_ENGLISH = {v: k for k, v in EMOJI_UNICODE_ENGLISH.items()}


@dataclasses.dataclass(frozen=True)
class VimGlobalStatus:
    cursor_colnum: int = (
        1  # 1 based. The first element of the line is 1. Even without lines.
    )
    cursor_lnum: int = (
        1  # 1 based. The first line is 1. Even without any char in the line.
    )
    cursor_want: int = 1  # Same as columns.

    screen_rows: int = 2
    screen_cols: int = 2
    screen_xpos: int = 0
    screen_ypos: int = 0

    buffer: List[str] = dataclasses.field(default_factory=list)
    # paste: str = ""
    # command: str = ""
    # search: str = ""
    command: List[str] = dataclasses.field(default_factory=list)

    should_quit: bool = False

    # action: None = None  # Placeholder for any other kind of action outside the status

    def copy(self, **changes):
        return dataclasses.replace(self, **changes)

    def with_added_command(self, command: List[str]):
        return self.copy(command=self.command + command)

    allowed_commands = "ALL"


@enum.unique
class SpecialChar(enum.Enum):
    BRICKS = EMOJI_UNICODE_ENGLISH[":brick:"]  # 🧱

    @staticmethod
    def is_emoji(s):
        return s in UNICODE_EMOJI_ALIAS_ENGLISH


END_OF_LINE_MAX = 2147483647

# Lines to keep between the cursors and the beginning/end of the screen.
# This doesn't work if it's the actual beginning/end of the file
LINE_BUFFER = 5



import dataclasses
import json
import subprocess
import tempfile
from functools import cached_property
from pathlib import Path
from shutil import which
from typing import List

from vim_quest_cli.deps.is_js import IS_JS

if not IS_JS:
    import curses.ascii
    import importlib_resources


from vim_quest_cli.engine.engine_interface import (
    EngineInterface,
    EngineState,
    CursorPos,
)

from vim_quest_cli.engine.vim_exe import resources as vim_resources


class Constants:
    @cached_property
    def SAVE_STATE_VIM(self):
        return (
            importlib_resources.files(vim_resources)
            .joinpath("save_state.vim")
            .read_text()
        )


CONSTANTS = Constants()


VIM_COMMAND = "vim"


def command_interpret(cmd):
    return cmd.replace("<ESC>", chr(curses.ascii.ESC))


class VimNotExistsException(Exception):
    pass


@dataclasses.dataclass(frozen=True)
class VimExeCursorPosition:
    lnum: int
    bufnum: int
    col: int
    off: int
    curswant: int


@dataclasses.dataclass(frozen=True)
class VimStateResults:
    command: str  # Same as input, can be ignored
    buffer: List[str]
    cursor: VimExeCursorPosition

    @classmethod
    def from_json(cls, j):
        return cls(
            command=j["command"],
            buffer=j["buffer"],
            cursor=VimExeCursorPosition(**j["cursor"]),
        )


def _vim_execute_actions(buffer: List[str], commands: List[str]) -> VimStateResults:
    if not which(VIM_COMMAND):
        raise VimNotExistsException()

    with tempfile.TemporaryDirectory() as tmp_path_str:
        tmp_path = Path(tmp_path_str)

        input_path = tmp_path / "input.txt"
        command_path = tmp_path / "commands.str"
        output_path = tmp_path / "output.json"
        vim_code_path = tmp_path / "save_state.vim"

        input_path.write_text("\n".join(buffer))
        command_path.write_text("".join(commands))
        vim_code_path.write_text(CONSTANTS.SAVE_STATE_VIM)

        cmd = [
            "vim",
            f"+:source {vim_code_path}",
            f'+call ExecuteAndSave("{input_path}", "{command_path}", "{output_path}")',
            "--not-a-term",
            "--noplugin",
        ]

        _ = subprocess.check_output(cmd, cwd=tmp_path_str)

        output_content = output_path.read_text()
        json_res = json.loads(output_content)
        print(json_res)
        return VimStateResults.from_json(json_res)


class VimExeEngine(EngineInterface):
    def process(self, state: EngineState) -> EngineState:
        res = _vim_execute_actions(state.buffer, state.command)

        return EngineState(
            cursor=CursorPos(
                line=res.cursor.lnum,
                col=res.cursor.col,
                col_want=res.cursor.curswant,
            ),
            buffer=res.buffer,
            command=[],
        )


__all__ = ["VimExeEngine"]

if __name__ == "__main__":
    print(_vim_execute_actions("hello world", "yyx"))
    print(
        _vim_execute_actions(
            "hello world", command_interpret(r"2litoto<ESC>Atiti<ESC>")
        )
    )



import dataclasses
from typing import Iterable

from vim_quest_cli.engine.engine_interface import EngineState, CursorPos
from vim_quest_cli.mode.mode_interface import Mode
from vim_quest_cli.view.view import ViewData, FgColor, CharacterView

import random


class GoToXMode(Mode):
    def _reset_path(self, state: EngineState, reset_cursor=False) -> EngineState:
        # First empty the state.
        nb_lines, nb_cols = 15, 40
        new_buffer = [["."] * nb_cols for _ in range(nb_lines)]
        new_buffer[random.randint(0, nb_lines - 1)][
            random.randint(0, nb_cols - 1)
        ] = "X"
        new_buffer[:] = ["".join(line) for line in new_buffer]

        cursor = state.cursor
        if reset_cursor:
            cursor = CursorPos(
                line=5,
                col=10,
                col_want=10,
            )

        return state.copy(buffer=new_buffer, cursor=cursor)

    def _init_after_params(self):
        self._state = self._reset_path(self._state, reset_cursor=True)

    @staticmethod
    def _char_under_cursor(state: EngineState):
        return state.buffer[state.cursor.line - 1][state.cursor.col - 1]

    def state_change(
        self, state_init: EngineState, state_end: EngineState
    ) -> EngineState:
        if GoToXMode._char_under_cursor(state_end) == "X":
            return self._reset_path(state_end)
        return state_end

    def _set_view_char(self, view: ViewData, char: str, line: int, column: int):
        # TODO: make sure the cursor and the line/col are on the same base.
        # I think it's going to be easier to be all 0 based from now on, and to adapt the interface to be otherwise.
        if line < 1 or line > len(view.chars):
            return
        if column < 1 or column > len(view.chars[0]):
            return
        current_val = view.chars[line - 1][column - 1]
        if current_val.ascii == "X":
            return
        view.chars[line - 1][column - 1] = CharacterView(
            ascii=char,
        )

    def change_view(self, view: ViewData):
        for line in view.chars:
            for character in line:
                if character.ascii == "X":
                    character.fg_color = FgColor.red
                if character.ascii == "#":
                    character.fg_color = FgColor.green

        self._set_view_char(view, "k", view.cursor.line - 1, view.cursor.column)
        self._set_view_char(view, "^", view.cursor.line - 2, view.cursor.column)
        self._set_view_char(view, "j", view.cursor.line + 1, view.cursor.column)
        self._set_view_char(view, "v", view.cursor.line + 2, view.cursor.column)
        self._set_view_char(view, "l", view.cursor.line, view.cursor.column + 2)
        self._set_view_char(view, ">", view.cursor.line, view.cursor.column + 4)
        self._set_view_char(view, "h", view.cursor.line, view.cursor.column - 2)
        self._set_view_char(view, "<", view.cursor.line, view.cursor.column - 4)

        return view



# pylint: disable=C0103

from ansi.sequence import sequence


# Cursor movement
up = sequence("A")
down = sequence("B")
forward = sequence("C")
back = sequence("D")
next_line = sequence("E")
prev_line = sequence("F")
goto_x = sequence("G")
goto = sequence("H", 2)
erase = sequence("J")
erase_data = erase
erase_line = sequence("K")
scroll_up = sequence("S")
scroll_down = sequence("T")
save_cursor = sequence("s")
load_cursor = sequence("u")
show = sequence("?25h")
hide = sequence("?25l")



# TODO: using .executor_interface instead of full path failed when bundle.
from contextlib import suppress
from typing import Optional

from vim_quest_cli.engine.engine_interface import (
    EngineInterface,
    EngineState,
    CursorPos,
)
from vim_quest_cli.interface.mapping import KeyNames


class NoCommandLeftException(Exception):
    pass


def clamp(n, minn, maxn):
    return max(min(maxn, n), minn)


NUMBERS = {0, 1, 2, 3, 4, 5, 6, 7, 9}
END_OF_LINE_MAX = 2147483647
LINE_BUFFER = 5


class PurePythonEngine(EngineInterface):
    def process(self, state: EngineState) -> EngineState:

        previous_state, next_state = state, state
        with suppress(NoCommandLeftException):
            while previous_state.command:
                next_state = self._execute_single_command(previous_state)
                previous_state = next_state

        return next_state

    def _execute_single_command(self, state: EngineState) -> EngineState:
        if not state.command:
            return state
        c = state.command[:]

        # First try to read numbers
        number = None
        if c[0] in [1, 2, 3, 4, 5, 6, 7, 8, 9]:
            number_str = c.pop(0)
            while c[0] in self.NUMBERS:
                number_str += c.pop(0)
            number = int(number_str)

        if c[0] == KeyNames.DOWN or c[0] == "j":
            return self._cursor_down(number, state.copy(command=c[1:]))

        if c[0] == KeyNames.UP or state.command[0] == "k":
            return self._cursor_up(number, state.copy(command=c[1:]))

        if c[0] == KeyNames.LEFT or state.command[0] == "h":
            return self._cursor_left(number, state.copy(command=c[1:]))

        if c[0] == KeyNames.RIGHT or state.command[0] == "l":
            return self._cursor_right(number, state.copy(command=c[1:]))

        if c[0] == "$":
            return self._cursor_end_of_line(number, state.copy(command=c[1:]))

        if c[0] == "0":
            return self._cursor_beginning_of_line(number, state.copy(command=c[1:]))

        if c[0] == "G":
            return self._cursor_end_of_file(number, state.copy(command=c[1:]))

        if c[0:2] == ["g", "g"]:
            return self._cursor_beginning_of_file(number, state.copy(command=c[2:]))

        if c[0:2] == ["d", "d"]:
            return self._delete_line(number, state.copy(command=c[2:]))

        if c[-1] == KeyNames.ESCAPE:  # TODO: it's not specifically like that
            return state.copy(command=[])

        raise NoCommandLeftException()

    @staticmethod
    def _cursor_down(number: Optional[int], state: EngineState) -> EngineState:
        number = number or 1
        new_lnum = min(state.cursor.line + number, len(state.buffer))
        col = min(
            state.cursor.col_want, PurePythonEngine._get_line_len(state, new_lnum)
        )
        return state.copy(cursor=state.cursor.copy(line=new_lnum, col=col))

    @staticmethod
    def _cursor_up(number: Optional[int], state: EngineState) -> EngineState:
        number = number or 1
        new_lnum = max(state.cursor.line - number, 1)
        col = min(
            state.cursor.col_want, PurePythonEngine._get_line_len(state, new_lnum)
        )
        return state.copy(cursor=state.cursor.copy(line=new_lnum, col=col))

    @staticmethod
    def _cursor_left(number: Optional[int], state: EngineState) -> EngineState:
        number = number or 1
        new_col = max(state.cursor.col - number, 1)
        return state.copy(cursor=state.cursor.copy(col=new_col, col_want=new_col))

    @staticmethod
    def _cursor_right(number: Optional[int], state: EngineState) -> EngineState:
        number = number or 1
        new_col = min(state.cursor.col + number, PurePythonEngine._get_line_len(state))
        return state.copy(cursor=state.cursor.copy(col=new_col, col_want=new_col))

    @staticmethod
    def _cursor_end_of_line(number: Optional[int], state: EngineState) -> EngineState:
        number = number or 1
        new_lnum = min(state.cursor.line + number - 1, len(state.buffer))
        if number == 1:  # Normal case
            return state.copy(
                cursor=state.cursor.copy(
                    line=new_lnum,
                    col=PurePythonEngine._get_line_len(state, new_lnum),
                    col_want=END_OF_LINE_MAX,
                )
            )

    @staticmethod
    def _cursor_beginning_of_line(
        number: Optional[int], state: EngineState
    ) -> EngineState:
        if number is not None:
            raise ValueError(
                "Cannot have beginning of line with number != 1 \n\n" + repr(state)
            )
        return state.copy(
            cursor=state.cursor.copy(
                col=1,
                col_want=1,
            )
        )

    @staticmethod
    def _cursor_end_of_file(number: Optional[int], state: EngineState) -> EngineState:
        if number is None:
            number = len(state.buffer)
        # TODO : It starts at the first non-space character.
        new_lnum = clamp(number, 0, len(state.buffer))
        new_col_num = PurePythonEngine._get_first_non_space_char(state, new_lnum)
        return state.copy(
            cursor=state.cursor.copy(
                line=new_lnum,
                col=new_col_num,
                col_want=new_col_num,
            )
        )

    @staticmethod
    def _cursor_beginning_of_file(
        number: Optional[int], state: EngineState
    ) -> EngineState:
        if number is None:
            number = 1
        new_lnum = clamp(number, 0, len(state.buffer))
        new_col_num = PurePythonEngine._get_first_non_space_char(state, new_lnum)
        return state.copy(
            cursor=state.cursor.copy(
                line=new_lnum,
                col=new_col_num,
                col_want=new_col_num,
            )
        )

    @staticmethod
    def _get_line_len(
        state: EngineState, line: Optional[int] = None, min_1: bool = True
    ) -> int:
        lnum = state.cursor.line if line is None else line
        line_len = len(state.buffer[lnum - 1])
        if min_1:
            line_len = max(line_len, 1)
        return line_len

    @staticmethod
    def _get_first_non_space_char(
        state: EngineState, lnum: Optional[int] = None, min_1: bool = True
    ) -> int:
        """Used for cursor positioning"""
        lnum = state.cursor.line if lnum is None else lnum
        line = state.buffer[lnum - 1]
        res = 0
        for res, char in enumerate(line):
            if char not in (" ", "\t"):
                break
        return max(res + 1, 1 if min_1 else 0)

    @staticmethod
    def _refresh_view(state):
        cursor_y = state.cursor_lnum

        # Find the max and min of the view, count the lines buffers and absolute min max
        max_view = min(
            cursor_y - 1 - LINE_BUFFER, len(state.buffer) - state.screen_rows - 1
        )
        min_view = max(cursor_y - state.screen_rows + LINE_BUFFER, 0)

        # If the view is inside
        if state.screen_ypos <= min_view:
            return state.copy(screen_ypos=min_view)

        if state.screen_ypos >= max_view:
            return state.copy(screen_ypos=max_view)

        return state

    def _delete_line(self, number: Optional[int], state: EngineState):
        new_buffer = (
            state.buffer[0 : state.cursor.line] + state.buffer[state.cursor.line + 1 :]
        )
        return state.copy(buffer=new_buffer)



"""
Map between code received and a text representing the code launched.
"""
import enum
import string

OUTSIDE_OF_ENUM = set(string.ascii_lowercase + string.ascii_uppercase + string.digits)


class KeyNames(enum.Enum):
    CTRL_C = "#ctrl-c"
    LEFT = "#left"
    RIGHT = "#right"
    UP = "#up"
    DOWN = "#down"
    ESCAPE = "#escape"


CODE_2_KEY = {
    "\x03": KeyNames.CTRL_C,
    "\x1b": KeyNames.ESCAPE,
    "\x1b[D": KeyNames.LEFT,
    "\x1b[A": KeyNames.UP,
    "\x1c[C": KeyNames.RIGHT,
    "\x1b[B": KeyNames.DOWN,
}


class UnknownKeyMapping(Exception):
    pass


def input_code_to_enum(input_code):
    if input_code in OUTSIDE_OF_ENUM:
        return input_code
    if input_code in CODE_2_KEY:
        return CODE_2_KEY[input_code]
    raise UnknownKeyMapping(input_code)



import sys

from typing import Any, List, Optional, TYPE_CHECKING, Union

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

if TYPE_CHECKING:
    from ansi.colour.base import Graphic

CSI = "\x1b["
OSC = "\x1b]"
BEL = "\a"


class SequenceGenerator(Protocol):
    def __call__(self, *vals: Union[int, str, "Graphic"]) -> str:
        ...


def sequence(
    letter: str,
    fields: int = 1,
    default: Optional[List[Union[int, str, "Graphic"]]] = None,
) -> SequenceGenerator:
    """Create a function that creates escape sequences."""

    def _sequence(*values: Union[int, str, "Graphic"]) -> str:
        output = list(values)
        if fields >= 0 and len(output) > fields:
            raise ValueError(
                "Invalid number of fields, got %d expected %d" % (len(output), fields)
            )

        if default is not None:
            output = output + default[-(fields - len(output)) :]

        return "".join(
            [
                CSI,
                ";".join(map(str, output)),
                letter,
            ]
        )

    return _sequence


class OSCGenerator(Protocol):
    def __call__(*args: str, **kwargs: Any) -> str:
        ...


def osc(number: Union[int, str]) -> OSCGenerator:
    """Create a function that creates OSC sequences."""

    def _osc(*args: str, **kwargs: Any) -> str:
        key_value_pairs = ["%s=%s" % (k, v) for (k, v) in kwargs.items()]
        return "".join(
            [
                OSC,
                str(number),
                ";",
                ";".join(list(args) + key_value_pairs),
                BEL,
            ]
        )

    return _osc



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vim_quest_cli',
 'vim_quest_cli.corpus',
 'vim_quest_cli.deps',
 'vim_quest_cli.deps.ansi',
 'vim_quest_cli.deps.ansi.colour',
 'vim_quest_cli.engine',
 'vim_quest_cli.engine.pure_python',
 'vim_quest_cli.engine.vim_exe',
 'vim_quest_cli.engine.vim_exe.resources',
 'vim_quest_cli.interface',
 'vim_quest_cli.mode',
 'vim_quest_cli.old',
 'vim_quest_cli.view']

package_data = \
{'': ['*']}

install_requires = \
['ansi>=0.3.6,<0.4.0', 'argparse-dataclass>=0.2.1,<0.3.0']

entry_points = \
{'console_scripts': ['main = vim_quest_cli.__main__:main',
                     'stdin_test = vim_quest_cli.deps.test_stdin:main']}

setup_kwargs = {
    'name': 'vim-quest-cli',
    'version': '0.1.5',
    'description': '',
    'long_description': '# VimQuestCLI\n\n\nThis is a vim implementation in python that tries to implement game using the vim shortcuts.\n\n## Why python ?\n\nI love python. Maybe one day it\'s going to be converted in Rust (for easy webassembly and faster execution). But for now it\'s python 3.8.\n\n## Choice when doing executions\n\nExecutor :\n- local (the python vim implementation)\n- vim (calling vim to execute stuff, for compatibility tests)\n\nTerminal handling :\n- Curses wrapper : the way it\'s usually done in python.\n  It allows for curses to catch the input.\n- "manual" curses : Without using the full curses capabilities.\n\nExecution loop :\n- One method to start the whole loop.\n- One method that have the input as argument and returns what to print.\n\nThe python version doesn\'t have just vim capabilities, but allows to expand for allowing for hooks and data attached to the cells.\n\n## Todo\n\n- Use micro-python for testing the front end side.\n- Separate presentation layer\n\n- Add a vim executor (to be able to compare tests).\n- Have a context that is not a copy but modify itself (for speed).\n- Have a bunch of hooks, and use them to have game modes.\n- Have a command to choose between game modes.\n- See how neovim is made so I can use the same architecture.\n- Syntax highlighting for python.\n- Maybe have syntax highliting for game modes too ?\n- Solve the ANSI speed.\n- Activate the UTF-8, see how I can activate it.\n- Create docker images for website and terminal.\n- Add git hook for cleaning python before commit.\n- Use unicode characters (not necesseraly emojii) with colors for differents things happening.\n\nBUG CORRECTION :\n- Clean codebase.\n- ANSI is blinking.\n- Image location is not working.\n- Size of the terminal is not used.\n- Unicode block behavior is inconsistent.\n  - Maybe having replacement for unicode that include colors.\n\n# References\n\n- [vim-adventures.com](https://vim-adventures.com/) : really cool game. I keep paying for it as the licence is only for 6 months.\n- [vim.so](https://www.vim.so/lessons) : Manual tests using a modified javascript editor.\n- [vimgenius.com](http://www.vimgenius.com/lessons/vim-intro/levels/level-1)\n- [github.com/iggredible/Learn-Vim](https://github.com/iggredible/Learn-Vim/blob/master)\n\n- Python prompt toolking has a python implementation of vi commands :\n  - https://github.com/prompt-toolkit/python-prompt-toolkit/blob/master/src/prompt_toolkit/key_binding/bindings/vi.py\n\nNow vim implementationthat I could use for corectness :\n- ace js : https://github.com/ajaxorg/ace/blob/master/src/keyboard/vim_test.js\n- ideavim : https://github.com/JetBrains/ideavim/tree/master/src/test/java/org/jetbrains/plugins/ideavim\n\nncurses implementation : https://github.com/mscdex/node-ncurses\nMaybe this implementation would be easier to understand than the c ncurses one.\n# Next steps to cover\n\n- hjkl x\n- [number]command\n- i\n- aA\n- dd\n- wbe\n- v, Shift-v\n- ypP\n- "ci<symbol>"\n\n\n# Next step.\n\nVim core. That execute the commands and return a status. From a Contains the screen size and \n\nVim window. That take that buffer and return a screen. (matrix of data points with metadata).\n\nAttachable are game modes.\n\nThen there is terminal that push characters and returns a  \n\n\n\n\n\n',
    'author': 'Stolati',
    'author_email': 'stolati@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3.8
import argparse
import dataclasses
import enum

from vim_quest_cli.deps import create_modules_if_is_js  # Keep as the first module
from vim_quest_cli.corpus import labyrinth
from vim_quest_cli.engine.engine_interface import EngineState, EngineInterface
from vim_quest_cli.engine.pure_python.pure_python_engine import PurePythonEngine
from vim_quest_cli.engine.vim_exe.vim_exe_engine import VimExeEngine
from vim_quest_cli.interface.curses_window import Window
from vim_quest_cli.interface.terminal_ui import TerminalUI
from vim_quest_cli.mode.go_to_x import GoToXMode
from vim_quest_cli.mode.labyrinth import LabyrinthMode
from vim_quest_cli.mode.mode_interface import Mode
from vim_quest_cli.view.view import ViewFactory


class Engines(enum.Enum):
    PURE_PYTHON = PurePythonEngine
    VIM_EXE = VimExeEngine
    DEFAULT = PurePythonEngine


class Modes(enum.Enum):
    GOT_TO_X = GoToXMode
    LABYRINTH = LabyrinthMode


@dataclasses.dataclass(frozen=True)
class Options:
    engine_cls: EngineInterface
    terminal: str


def parse_args() -> Options:
    parser = argparse.ArgumentParser(description="Vim powaa.")
    parser.add_argument(
        "-e",
        "--engine",
        default=Engines.DEFAULT.value,
        type=lambda v: Engines[v].value,
        choices=[e.name for e in Engines],
        help="Select and engine to use",
    )
    parser.add_argument(
        "-t",
        "--terminal",
        default="ansi",
        choices=("ansi", "curses"),
        help="wich terminal to use",
    )
    # TODO add param for the mode
    # TODO add param for a file to load

    parsed_args = parser.parse_args()
    return Options(
        engine_cls=parsed_args.engine,
        terminal=parsed_args.terminal,
    )


def main():
    args = parse_args()

    state = EngineState(
        buffer="tu to\nto#to\nti#ti\ntu#tu\nti tu".split("\n"),
    )

    mode_cls = Modes.GOT_TO_X.value

    mode = mode_cls(
        engine=args.engine_cls(),
        view=ViewFactory(),
        state=state,
    )

    t = TerminalUI(mode=mode)

    if args.terminal == "ansi":
        t.start_ansi()
    elif args.terminal == "curses":
        t.start_curses()
    else:
        raise ValueError("Terminal parameters not in [ansi|curses]")


def as_single_method():
    t = TerminalUI(
        mode=Modes.GOT_TO_X.value(
            engine=PurePythonEngine(),
            view=ViewFactory(),
            state=EngineState(
                buffer="tu to\nto#to\nti#ti\ntu#tu\nti tu".split("\n"),
            ),
        ),
    )

    def res_method(user_input: str) -> str:
        res = []
        w = Window(stdout_write_fct=res.append)
        t.loop_one_step(user_input, w)

        return "".join(res)

    return res_method


if __name__ == "__main__":
    main()



""" Transform an Engine state into a matrix of chars. """
import dataclasses
import enum
from typing import List, Optional
from ansi.colour.base import Graphic

from vim_quest_cli.engine.engine_interface import EngineState

RESET_STYLE = str(Graphic("0"))


class FgColor(enum.Enum):
    # TODO use ansi.color.fg instead
    black = str(Graphic("30"))
    red = str(Graphic("31"))
    green = str(Graphic("32"))
    yellow = str(Graphic("33"))
    blue = str(Graphic("34"))
    magenta = str(Graphic("35"))
    cyan = str(Graphic("36"))
    white = str(Graphic("37"))


class BgColor(enum.Enum):
    black = str(Graphic("40"))
    red = str(Graphic("41"))
    green = str(Graphic("42"))
    yellow = str(Graphic("43"))
    blue = str(Graphic("44"))
    magenta = str(Graphic("45"))
    cyan = str(Graphic("46"))
    white = str(Graphic("47"))


@dataclasses.dataclass()
class CharacterView:
    ascii: str = " "  # Should always be of length 1
    fg_color: Optional[FgColor] = None
    bg_color: Optional[BgColor] = None


@dataclasses.dataclass()
class CursorPos:
    line: int = 1
    column: int = 1


@dataclasses.dataclass()
class ViewData:
    chars: List[List[CharacterView]] = dataclasses.field(default_factory=list)
    cursor: CursorPos = CursorPos()


class ViewFactory:
    @staticmethod
    def createView(engine_state: EngineState):
        return ViewData(
            chars=[[CharacterView(c) for c in line] for line in engine_state.buffer],
            cursor=CursorPos(
                line=engine_state.cursor.line, column=engine_state.cursor.col
            ),
        )



# TODO: tests that take a mode (every mode with seed needs a seed in the init)
# and do unittest on them. Pushing a few key and expecting the output.
from typing import Iterable

from vim_quest_cli.engine.engine_interface import EngineInterface, EngineState
from vim_quest_cli.view.view import ViewFactory, ViewData


class Mode:
    engine: EngineInterface
    state: EngineState
    view: ViewFactory

    def __init__(self, engine: EngineInterface, state: EngineState, view: ViewFactory):
        self._engine = engine
        self._state = state
        self._view = view
        self._init_after_params()

    # TODO : change str to an enum there
    def feedkey(self, keys: Iterable[str] = ()) -> ViewData:

        state_previous = self._state.with_added_command(keys)
        state_next = self._engine.process(self._state.with_added_command(keys))

        self._state = self.state_change(state_previous, state_next)

        view = self._view.createView(self._state)
        return self.change_view(view)

    # To be overloaded by subclasses
    def _init_after_params(self):
        pass

    def change_view(self, view: ViewData) -> ViewData:
        return view

    def state_change(
        self, state_init: EngineState, state_end: EngineState
    ) -> EngineState:
        return state_end


"""

the screeze has to be passed around for mode to be able to use it.
Maybe in a GeneralConfig structure ?
- If the printing is unicode friendly or not.

    create_new_path
        - keep the curspor pos.
        - Generate a target to go
    When someting moves :
        - Check if the move ends up on the target,
        - If yes, then create new path.
    When printing happens :
        - change the target in red (first by just changing the target text by the red addition)

Improvements :
    - Cancel every non-movement.
    - Have the redness be part of the visual buffer.
    - Then add the target as metadata of the input, and use that to check what's happening.
"""

"""         
            k 
            ^ 
        h < x > l
            v
            j
             
"""


class PassthroughMode:

    pass



LORUM_TEXT = """
Lorem ipsum dolor sit amet
Curbatitur sapien neque

id nisl. Fusce id diam id
-
vitae mi non orci ultricies
""".split(
    "\n"
)

BLOCK_TEXT = r"""
 <- Cursor, move with 'j' 🧱.
                          🧱.
                          🧱.
                          🧱.
                          🧱.
                          🧱.
🧱🧱🧱🧱🧱🧱             🧱.
Blocked ?. Try 'l'          .
                            .
                            .
                            .
                            .
                            .
                            .
                            .
                            .
              🧱🧱🧱🧱🧱🧱 .
      Again ? Maybe with 'h'.
                            .
                            .
                            .
                            .
    🧱🧱🧱🧱🧱🧱🧱🧱🧱🧱🧱.
    🧱                      .
    🧱   ^     🧱           .
    🧱   |     🧱           .
    🧱   k     🧱           .
               🧱  goal->  🍩.
""".split(
    "\n"
)[
    1:-1
]



# Coming from https://github.com/stolati/pyProject2singleFile/blob/main/project2singleFile/export.py
import sys
from importlib import _bootstrap
from importlib.abc import MetaPathFinder
from types import ModuleType
from typing import Optional, Dict

from dataclasses import dataclass


@dataclass
class ModuleDef:
    name: str
    source: str
    is_package: bool
    file: str


ModuleDefsDict = Dict[str, ModuleDef]


class ModuleImporter(MetaPathFinder):
    _modules_defs: ModuleDefsDict

    @classmethod
    def install(cls, modules: ModuleDefsDict):
        sys.meta_path.insert(0, cls(modules))

    def __init__(self, modules_defs: ModuleDefsDict):
        self._modules_defs = modules_defs

    # Finder part

    def find_spec(self, fullname: str, path, target) -> Optional[_bootstrap.ModuleSpec]:
        if fullname not in self._modules_defs:
            return None

        module_def = self._modules_defs[fullname]
        spec = _bootstrap.spec_from_loader(fullname, self, origin="bundle_lib")
        spec._def = module_def
        return spec

    # Loader part

    def create_module(self, spec):
        spec._code = self.get_code(spec._def.name)

        module = ModuleType(name=spec.name)
        module.__package__ = module.__name__ = spec._def.name
        module.__file__ = spec._def.file
        module.__loader__ = self
        module.__spec__ = spec
        return module

    def exec_module(self, module=None):
        exec(module.__spec__._code, module.__dict__)

    def get_code(self, fullname):
        file = self._modules_defs[fullname].file
        return compile(self.get_source(fullname), file, "exec")

    def get_source(self, fullname):
        return self._modules_defs[fullname].source or ""

    def is_package(self, fullname):
        return self._modules_defs[fullname].is_package
